open Atoms ;;
open Terms ;;
open Substitutions ;;
open Multiset ;;

type border = unit term;;


let double_map (f:'a -> 'b -> 'c) (l1: 'a list) (l2: 'b list) =
  let rec apply_one res a = function
    | [] -> res
    | b::l -> apply_one ((f a b)::res) a l
  in let rec apply_two res l1 l2 =
    match l1 with
    | [] -> res
    | a::l -> apply_two (apply_one res a l2) l l2
  in apply_two [] l1 l2
;;


let develop alist =
  (*  'a list array --> 'a array list *)
  let copy_except_i arr a i =
    let arr2 = Array.make (Array.length arr) arr.(0) in
    let rec copy = function
      | j when j >= Array.length arr -> ()
      | j when j = i  -> arr2.(i) <- a ; copy (j+1)
      | j -> arr2.(j) <- arr.(j) ; copy (j+1)
    in copy 0 ; arr2

  in let rec on_ith i res arr =
    let l = arr.(i) in
    let rec aux res = function
      | [] -> res 
      | a::l -> aux ((copy_except_i arr [a] i)::res) l
    in aux res l
  
  in let ith i = 
    List.fold_left (on_ith i) [] 

  in let rec all_i res = function
    | n when n = Array.length alist -> res
    | i -> all_i (ith i res) (i+1)

  in let nettoie = function
    | [] -> assert(false)
    | a::t -> a

  in List.map (Array.map nettoie) (all_i [alist] 0)
;;


let tuple_map alist =
  let l = develop alist in
  List.map (fun a -> Tuple(a)) l
;;


let rec apply t a =
  match t with
  | At(_) -> assert(false) 
  | Omega -> assert(false) 
  (* On ne devrait pas appeler
   * la fonction dans ce cas.
   *)
  | Senc(_,_) ->
      double_map (fun u -> (fun v -> Senc(u,v))) a.(0) a.(1)
  | Aenc(_,_) ->
      double_map (fun u -> (fun v -> Aenc(u,v))) a.(0) a.(1)
  | Sign(_,_) ->
      double_map (fun u -> (fun v -> Sign(u,v))) a.(0) a.(1)
  | Hash(_)   -> List.map (fun f -> Hash(f)) a.(0)
  | Vk(_)     -> List.map (fun f -> Vk  (f)) a.(0)
  | Pub(_)    -> List.map (fun f -> Pub (f)) a.(0)
  | Tuple(t)  -> tuple_map a 
;;

(* border_type s'applique uniquement aux types.
 * true signifie qu'on est à l'intérieur
 * du type et que l'on doit donc faire attention
 * aux instantiations de variables. *)

let rec  border_type  = function
  | Omega -> assert(false)
  (* Omega ne devrait pas apparaître dans les règles *)
  | At(_) ->  [At()]
  | t -> let a = Array.map border_type (sons t) in
   (At() :: (apply t a))
;;


let rec border_term = function
  |Omega -> assert(false)
  | At(_,t)     -> border_type t
  | Tuple(t)    -> let a = Array.map border_term t in
                   tuple_map a
  | t -> let a = Array.map border_term (sons t) in
  (At() :: (apply t a))
;;

let borders ik_left = 
  develop (Array.map border_term ik_left);;

(* Le calcul des frontières est fini.
 * On passe au calcul des règles correspondantes *)

