open Atoms ;;
open Terms ;;
open Substitutions ;;
open Multiset ;;
open Preflat ;;


(* Fonctions utiles. *)
let rec get_iks_and_sts iks sts  = function
  | [] -> iks,sts
  | (State (c,sn,tp,tq))::l -> get_iks_and_sts iks (State (c,sn,tp,tq)::sts) l
  | (Phase n)::l -> get_iks_and_sts iks (Phase(n)::sts) l
  | (Ik (xp,xq))::l -> get_iks_and_sts (Ik (xp,xq)::iks) sts l
  | _ -> raise (Invalid_argument "in get_iks_and_sts in on_each_rule in process.ml")
;;

let add_vars (tg:'a term array) (l: 'a term) =
  let longueur = Array.length tg in
  let initialise = function
    | n when n = longueur -> l
    | n -> tg.(n)
  in Array.init (longueur+1) initialise
;;

let rec add_vars_list (tg: atom term array) = function
    | [] -> tg
    | u::l -> add_vars_list (add_vars tg u) l
;;


let gives_asym_lab r_id vars_l = function
    |Ptcl(_,c,n,ta,tb)  -> Asym(r_id,c,n,ta, add_vars_list tb vars_l)
    |Asym(_,c,n,ta,tb)  -> Asym(r_id,c,n,ta, add_vars_list tb vars_l)
    | _ -> raise (Invalid_argument "in gives_asym_lab in on_each_rule in process.ml")
;;

let change_id rid = function
  | Ptcl(_,c,n,ta,tb) -> Ptcl(rid,c,n,ta,tb) 
  | Asym(_,c,n,ta,tb) -> Asym(rid,c,n,ta,tb) 
  | _ -> 
      failwith("This function should only be used with Ptcl and Asym.")

let rec search_vars_arr fv nfv res = function
    | At(Var(n)) when  fv <= n && n < nfv -> At(Var(n))::res
    | At(_) -> res
    | t -> 
        let st = sons t in
        Array.fold_left (search_vars_arr fv nfv) res st
;;

let get_right = function
  |Ik(a,b) -> b
  | _ -> failwith("This function should only be applied on Ik facts.")
;;

let decompose r fv fid =
  let (l,pre,add,del) = r in
  let ik_l,sts_l = get_iks_and_sts [] [] (Array.to_list pre) in
  let iks = Array.of_list ik_l in
  let sts = Array.of_list sts_l in
  let flats = get_all_flats_versions fv iks in

  let on_one_item fv rid (fidele,sigmaP,sigmaQ,new_ik,new_fv) =
    let new_pre = Array.append sts new_ik in
    match fidele with
    | true -> 
        let new_l = change_id rid l in
        max fv new_fv,
        rule_map_old (sigmaP,sigmaQ) (new_l,new_pre,add,del)
    | false ->
        let ik_r = Array.map get_right new_ik in
        let new_vars = 
          Array.fold_left (search_vars_arr fv new_fv) [] ik_r in
        max fv new_fv,
        (gives_asym_lab rid new_vars l,new_pre,[|Bad|],[||])
  
  in let rec aux fv rid res = function
    | [] -> res,fv,rid
    | a::l -> 
        let new_fv,new_r = on_one_item fv rid a in
      aux new_fv (rid+1) (new_r::res) l
  in aux fv fid [] flats

