open Translation ;; 
open Parser ;; 
open Type_compliant;;
open Static ;;
open Terminaison;;
open Algo;;

let version  = "SAT-Equiv version 0.3\n";;

exception Contradiction_between_arguments ;;
exception Without_file ;;
exception End ;;

let find_file_name a =
  let rec aux = function
    | n when n = Array.length a -> raise Without_file
    | n when not ((Str.string_before a.(n) 1) = "-") -> a.(n)
    | n -> aux (n+1)
  in aux 1
;;

let make_clean s =
  let slash = Str.regexp "/" in
  let list_words = Str.split slash s in
  let rec take_last = function
    | [] -> assert(false)
    |[a] -> a
    | a::t -> take_last t
  in take_last list_words 
;;


let main () =
        let argument_reader =
                let equiv       = ref true  in
                let pq          = ref true in (* true : p , false : q - inutile si !equiv = true *)
                let efface      = ref true in
                let static      = ref true in
                let onlyBound   = ref false in
                let print       = ref false in
                let compliance_check = ref "yes" in
                let constants = ref None in (* None : (c0,c1,Omega), Some(true) : (c0,Omega), Some(false): (c0,c1) *)
                let spec_list   =
                        [
                        ("--inclusionP" , Arg.Unit (fun () -> equiv := false; pq := true ) , "Tests only the inclusion of P in Q");
                        ("-p"           , Arg.Unit (fun () -> equiv := false; pq := true ) , "Same as --inclusionP");
                        ("--inclusionQ" , Arg.Unit (fun () -> equiv := false; pq := false ) , "Tests only the inclusion of Q in P");
                        ("-q"           , Arg.Unit (fun () -> equiv := false; pq := false ) , "Same as --inclusionQ");
                        ("--no-static"  , Arg.Unit (fun () -> static := false ) , "Does not test static inclusion.");
                        ("-s"           , Arg.Unit (fun () -> static := false ) , "Same as --no-static");
                        ("--print-graph", Arg.Unit (fun () -> print  := true  ) , "Print the planning graph.");
                        ("-g"           , Arg.Unit (fun () -> print  := true  ) , "Same as --print-graph");
                        ("--bound-only" , Arg.Unit (fun () -> onlyBound  := true  ) , "Only computes the bound.");
                        ("-b"           , Arg.Unit (fun () -> onlyBound  := true  ) , "Same as --bound-only");
                        ("--keep"       , Arg.Unit (fun () -> efface := false ) , "Keeps the last files used by the SAT solver");
                        ("-k"           , Arg.Unit (fun () -> efface := false ) , "Same as --keep");
                        ("--compliance"     , Arg.Unit (fun () -> () ) , "Checks the compliance and print a message if the protocol is not type-compliant (default)");
                        ("--compliance-only", Arg.Unit (fun () -> compliance_check := "only" ) , "Only checks compliance");
                        ("-o"               , Arg.Unit (fun () -> compliance_check := "only" ) , "Same as --compliance-only");
                        ("--no-compliance"  , Arg.Unit (fun () -> compliance_check := "none" ) , "Does not check compliance");
                        ("-n"               , Arg.Unit (fun () -> compliance_check := "none" ) , "Same as --no-compliance");
                        ("-w", Arg.Unit (fun () -> constants := Some(true ) ), "Uses only one constant plus Omega (c0,Omega)" );
                        (* w comme omega *)
                        ("-t", Arg.Unit (fun () -> constants := Some(false) ), "Uses only two constants (c0,c1)" );
                        (* t comme typed *)
                        ("--version", Arg.Unit (fun () -> print_string version ; raise End ), "Print version number" );
                        ("-v", Arg.Unit        (fun () -> print_string version ; raise End ), "Same as --version" );
                        ]
                in let usage_msg = "SatEquiv is an equivalence checker for cryptographic protocols.\n\n Usage: ./satequiv [options] file [options] \n Please do not use any symbol - (minus) at the beginning of your file name.\n\n Options available:\n"
                in Arg.parse spec_list print_endline usage_msg; (!equiv,!pq,!static,!onlyBound,!print,!efface,!compliance_check,!constants)


        in let argName = find_file_name Sys.argv
        in let file = String.concat "" ["__planning_";make_clean argName]
        in let file_sat = String.concat "." [file ; "dimacs"]
        in let file_out = String.concat "." [file ; "out"]
        in let instructions onlyBound print static assym priv twoConstants max_tuple p_include_q 
        ik_parse fresh_var =
                let our_facts,our_rules = 
                  get_all_rules twoConstants max_tuple p_include_q 
                  ik_parse fresh_var in
                
                let static_rules = 
                  all_static_rules (5+max_tuple) assym p_include_q in
                let b = borne p_include_q priv in

                match onlyBound with

                | true -> 
                    print_string "The bound is " ; 
                    print_int (b+2) ;
                    print_string ".\n";
                    true
                
                | false -> algo print static b file_sat file_out our_facts our_rules static_rules

                
        in let when_equiv onlyBound print static assym priv twoConstants compliance_check max_tuple p_include_q q_include_p ik_parse fresh_var =
                print_string "P include Q:\n" ;
                begin
                if compliance_check && not (check_compliance_biprotocol p_include_q) 
                then print_string "Warning: P is not type compliant!\n"
                end ;
                match instructions onlyBound print static assym priv twoConstants max_tuple p_include_q ik_parse fresh_var 
                with
                | true -> 
                    let _ =
                      print_string "Q include P:\n" ;
                      begin
                      if compliance_check && not (check_compliance_biprotocol q_include_p) 
                      then print_string "Warning: Q is not type compliant!\n"
                      end ;
                      instructions onlyBound print static assym priv twoConstants max_tuple q_include_p ik_parse fresh_var
                    in ()
                | false -> ()
                
        in let when_include onlyBound print static assym priv twoConstants pq compliance_check max_tuple p_include_q ik_parse fresh_var =
                
                let warningMessage,reminder = match pq with
                | true  -> "Warning: P is not type compliant!\n","P include Q:\n"
                | false -> "Warning: Q is not type compliant!\n","Q include P:\n"
                
                in print_string reminder ;
                begin
                if compliance_check && not (check_compliance_biprotocol p_include_q) 
                then print_string warningMessage ;
                let _ =  instructions onlyBound print static assym priv twoConstants max_tuple p_include_q ik_parse fresh_var
                in ()
                end


        in let (equiv,pq,static,onlyBound,print,efface,compliance_check,constants) = argument_reader
        in let ic_parse = 
                try open_in argName
                with
                | Invalid_argument _ (*"index out of bounds"*) -> raise Without_file
        in let assym,priv,ik_parse,max_tuple,p_include_q,q_include_p,fresh_var = complete_parser (not (constants = Some(false))) ic_parse
        in let inclusionProcess = 
                match pq with
                | true -> p_include_q
                | false -> q_include_p
        in begin match equiv,compliance_check with
        | true,"only"  -> 
                        let compliantp = check_compliance_biprotocol p_include_q in
                        let compliantq = check_compliance_biprotocol q_include_p in
                        begin
                                match compliantp,compliantq with
                                | true,true -> print_string "P and Q are type compliant.\n"
                                | true,false -> print_string "P is type compliant but Q is not (!).\n"
                                | false,true -> print_string "P is not type compliant (!) but Q is.\n"
                                | _ -> print_string "Neither P nor Q is type compliant !!\n"
                        end
        | false,"only" ->
                        let compliantp = check_compliance_biprotocol inclusionProcess in
                        begin
                                match compliantp,pq with
                                | true,true  -> print_string "P is type compliant.\n"
                                | true,false -> print_string "Q is type compliant.\n"
                                | false,true  -> print_string "P is not type compliant !\n"
                                | false,false -> print_string "Q is not type compliant !\n"
                        end
        | true, "none" -> when_equiv onlyBound print static assym priv (not (constants = Some(true))) false max_tuple p_include_q q_include_p ik_parse fresh_var
        | true, "yes"  -> when_equiv onlyBound print static assym priv (not (constants = Some(true))) true  max_tuple p_include_q q_include_p ik_parse fresh_var
        | false,"none" -> when_include onlyBound print static assym priv (not (constants = Some(true))) pq false max_tuple inclusionProcess ik_parse fresh_var
        | false,"yes"  -> when_include onlyBound print static assym priv (not (constants = Some(true))) pq true  max_tuple inclusionProcess ik_parse fresh_var
        | _ -> raise (Invalid_argument "impossible")
        end ;

        if efface then let _ = Unix.system (String.concat " " ["rm -f" ; file_sat ; file_out ] ) in ()
;;

try main () with
        | Without_file -> print_string "You should indicate a file or type ./satequiv --help for more informations.\n"  
        | Sys_error(_) -> print_string "There was a Sys_error. Please check the name of your file.\n"
        | End -> ()
