open Atoms ;;
open Terms ;;

type 'a process = 
  In of int*('a term)*('a process) 
| Out of int*('a term)*('a process) 
| PhaseRestriction of int*('a process) 
| Null ;;

type typed_process = (atom*complete_type) process ;;

type protocol = (typed_process * atom process) list ;;

let rec untype_process = function
        | Null -> Null
        | In(c,t,p) -> In(c,untype t,untype_process p)
        | Out(c,t,p) -> Out(c,untype t,untype_process p)
        | PhaseRestriction(n,p) -> PhaseRestriction(n,untype_process p) 
;;

let get_next = function
        | In(_,_,x) -> x
        | Out(_,_,x) -> x
        | PhaseRestriction(_,p) -> p 
        | Null -> raise (Invalid_argument "null process in get_next in process.ml")
;;

let rec get_channel = function
        | In(c,_,_) -> c
        | Out(c,_,_) -> c
        | PhaseRestriction(_,p) -> get_channel p
        | Null ->  raise (Invalid_argument "null process in get_channel in process.ml")
;;

let is_null = function
        | Null -> true
        | _ -> false
;;

let rec process_length (maxPh,resPh,resOut,resIn) = function
  | Null -> (maxPh,resPh,resOut,resIn)
  | In(_,_,p)  -> process_length (maxPh,resPh,resOut,resIn+1) p
  | Out(_,_,p) -> process_length (maxPh,resPh,resOut+1,resIn) p
  | PhaseRestriction(n,p) -> process_length (max n maxPh,1+resPh,resOut,resIn) p
;;

let max_interleaving_length (p:protocol) =
  let rec aux (maxPh,resPh,resOut,resIn) = function
  | [] -> (maxPh,resPh,resOut,resIn)
  | (a,_)::t -> aux (process_length (maxPh,resPh,resOut,resIn) a) t
  in aux (0,0,0,0) p
;;



let rec print_process print_term = function
        | Null -> print_string "0" 
        | In(n,t,p) -> 
                        begin 
                                print_string "in(c_" ; 
                                print_int n ; print_string ", " ; 
                                print_term t ; print_string ")." ; 
                                print_process print_term p
                        end
        | Out(n,t,p) ->
                        begin 
                                print_string "out(c_" ; 
                                print_int n ; print_string ", " ; 
                                print_term t ; print_string ")." ; 
                                print_process print_term p
                        end
        |PhaseRestriction(n,p) ->
            begin
              print_int n ;
              print_string ":" ;
              print_process print_term p
            end
;;

let rec max_tuple_size_process n = function
  | Null -> n
  | In(_,t,p)  -> max_tuple_size_process (max_tuple_size_term n t) p
  | Out(_,t,p) -> max_tuple_size_process (max_tuple_size_term n t) p
  | PhaseRestriction(n,p) -> max_tuple_size_process n p
;;

let rec max_tuple_size_typed_process n = function
  | Null -> n
  | In(_,t,p)  -> max_tuple_size_typed_process (max_tuple_size_typed n t) p
  | Out(_,t,p) -> max_tuple_size_typed_process (max_tuple_size_typed n t) p
  | PhaseRestriction(n,p) -> max_tuple_size_typed_process n p
;;

let max_tuple_size (p:protocol) =
  let rec aux n = function
    | [] -> n
    | (p,q)::t -> 
        let m = max_tuple_size_typed_process n p in
        aux (max_tuple_size_process m q) t
         
  in aux 0 p
;;

let print_process_P = print_process print_typed_term ;;
let print_process_Q = print_process print_term ;;

