open Atoms ;;
open Terms ;;
open Multiset ;;
open Process ;;
open Variables ;;
open Flattening ;;


let untype_array = Array.map untype ;;

let to_rule maxPh (p:typed_process) (q:atom process) (list_of_rules:rule list) (input_types:atom_type term list) fresh_var fresh_id =
        let rec dephase ph = function
          | PhaseRestriction(n,p) when n>=ph -> dephase n p
          | PhaseRestriction(_,_) (* n<ph *) -> (ph,Null)
          | p -> ph,p 
        in 
        let rec single_rule ph maxPh list_of_rules input_types c p0 q0 tp tq sn fresh_var fresh_id =
                match p0,q0 with
                |Out(cp,up,p1),Out(cq,uq,q1) ->
                                let tp1,tp_lab = variables_state_typed   tp (Out(cp,up,p1)) in
                                let tq1,tq_lab = variables_state_untyped tq (Out(cq,uq,q1)) in
                                let new_rule = 
                                        (Ptcl(fresh_id,c,sn,untype_array tp_lab,tq_lab) , 
                                        [|Phase ph ; State(c,sn,tp,tq)|] ,
                                        [|State(c,sn+1,tp1,tq1) ; Ik(up,uq) |],
                                        [|State(c,sn,tp,tq)|])
                                in principale ph ph maxPh (new_rule::list_of_rules) input_types c p1 q1 tp1 tq1 (sn+1) fresh_var (fresh_id+1)
                |In(cp,up,p1),In(cq,uq,q1) ->
                               let tp1,tp_lab = variables_state_typed   tp (In(cp,up,p1)) in
                               let tq1,tq_lab = variables_state_untyped tq (In(cq,uq,q1)) in
                               let new_rule =
                                       (Ptcl(fresh_id,c,sn,untype_array tp_lab,tq_lab) , 
                                       [|Phase ph ; State(c,sn,tp,tq); Ik(up,uq) |] ,
                                       [|State(c,sn+1,tp1,tq1)|],
                                       [|State(c,sn,tp,tq)|])
                               in let tp = get_type up 
                               in let decomposed_rules,nfv,nfid = decompose new_rule fresh_var (fresh_id+1)
                               in principale ph ph maxPh (List.rev_append decomposed_rules list_of_rules) (tp::input_types) c p1 q1 tp1 tq1 (sn+1) nfv nfid
               | In(cp,up,_),_ ->
                               let _,tp_lab = variables_state_typed   tp (In(cp,up,Null)) in
                               let _,tq_lab = variables_state_untyped tq (In(cp,x fresh_var,Null)) in
                               let new_rule =
                                       (Asym(fresh_id,c,sn,untype_array tp_lab,tq_lab) , 
                                       [|Phase ph ; State(c,sn,tp,tq); Ik(up,x fresh_var)|], 
                                       [|Bad |],
                                       [||])
                               in let tp = get_type up 
                               in let decomposed_rules,nfv,nfid = decompose new_rule (fresh_var+1) (fresh_id+1)
                               in maxPh,(List.rev_append decomposed_rules list_of_rules),(tp::input_types),nfv,nfid
               | _ -> raise (Invalid_argument "single_rule in to_rule in process.ml")
        
        and principale phP phQ maxPh list_of_rules input_types c p q tp tq sn fresh_var fresh_id =
          let (phP0,p0) = dephase phP p in
          let (phQ0,q0) = dephase phQ q in
          let maxPh0 = max phP0 maxPh in
            match p0,q0 with
               | (p1,_) when phQ0 != phP0 -> 
                   principale phP0 phP0 maxPh0 list_of_rules input_types c p1 Null tp tq sn fresh_var fresh_id
               | Null,_ -> maxPh0,list_of_rules,input_types,fresh_var,fresh_id
               | Out(cp,_,_),Out(cq,_,_) when ( c != cp || c != cq) -> raise (Invalid_argument "channel error in to_rule in process.ml")
               | Out(cp,up,p1),Out(cq,uq,q1) -> 
                               single_rule phP0 maxPh0 list_of_rules input_types c (Out(cp,up,p1)) (Out(cq,uq,q1)) tp tq sn fresh_var fresh_id
               | Out(_,up,_),_ ->
                               let new_rule =
                                       (Asym(fresh_id,c,sn,untype_array tp,tq) , 
                                       [|Phase phP0; State(c,sn,tp,tq)|] ,
                                       [|Bad ; Ik(up,At(Cst 0)) |],
                                       [||])
                               in maxPh0,(new_rule::list_of_rules),input_types,fresh_var,fresh_id+1
               | In(cp,_,_),In(cq,_,_) when ( c != cp || c != cq) -> 
                   raise (Invalid_argument "channel error in to_rule in process.ml")
               | In(cp,up,p1),In(cq,uq,q1) -> 
                   single_rule phP0 maxPh0 list_of_rules input_types c (In(cp,up,p1)) (In(cq,uq,q1)) tp tq sn fresh_var fresh_id
               | In(cp,up,p1),q -> 
                   single_rule phP0 maxPh0 list_of_rules input_types c (In(cp,up,p1)) q tp tq sn fresh_var fresh_id
               | _ -> raise (Invalid_argument "Impossible to get a PhaseRestriction at this point")
                               
        in match (is_null p) with
        | true -> maxPh,list_of_rules,input_types,fresh_var,fresh_id
        | false -> let c = get_channel p in principale 0 0 maxPh list_of_rules input_types c p q [||] [||] 0 fresh_var fresh_id
;;

let decomposition_rules max_tuple_size =
        (* Règles de déduction *)

  (* Dans size_n_tuple, le 5+n provient des
   * 3 règles de déduction
   * (sdec + adec + getmsg)
   * 2 règles pour l'équivalence statique 
   * (égalité + checksign)
   * Soit 5 règles de base.
   * Donc on commence à 6 pour les tuples
   *)
  let size_n_tuple n =
    let arrP = Array.init n (fun k -> yt k (At(Cameleon))) in
    let arrQ = Array.init n (fun k -> x (n+k)) in
    Adv(5+n,Array.map untype arrP,arrQ),
    [|Ik(Tuple(arrP),Tuple(arrQ))|],
    Array.init n (fun k -> Ik(arrP.(k),arrQ.(k))),
    [||]

  in let rec all_tuple_rules res = function
    | 0 -> res
    | 1 -> res (* inutile de décomposer des tuples de taille 1
                * car le parseur exclut ce cas *)
    | n -> all_tuple_rules (size_n_tuple n :: res) (n-1)
  in let base = 
        [
                Adv(0,[|x 0 ; x 1|],[|x 2 ; x 3|]),
                [|Ik(Senc(yt 0 (At(Cameleon)),yt 1 (At(Cameleon))) , Senc(x 2,x 3)) ; Ik(yt 1 (At(Cameleon)),x 3) |],
                [|Ik(yt 0 (At(Cameleon)),x 2)|],
                [||];

                Adv(1,[|x 0;x 1|],[|x 2;x 3|]),
                [|Ik(Aenc(yt 0 (At(Cameleon)),Pub( At(Var(1), At(Cameleon)))),Aenc(x 2,Pub(x 3)));Ik(At(Var(1),At(Cameleon)),x 3)|],
                [|Ik(yt 0 (At(Cameleon)),x 2)|],
                [||];

                Adv(2,[|x 0;x 1|],[|x 2;x 3|]),
                [|Ik(Sign(yt 0 (At(Cameleon)), At(Var(1), At(Cameleon))),Sign(x 2,x 3)) |],
                [|Ik(yt 0 (At(Cameleon)),x 2)|],
                [||];
]
  in all_tuple_rules base max_tuple_size
;;

(* Je suppose que le protocole arrive sous la forme suivante :
        * (P0,Q0),...(Pn,Qn)
        *) 
let rec add_phase_rules rules = function
  | 0 -> rules
  | n -> 
      let r =
        (PhaseIncr(n-1,n),[|Phase(n-1)|],[|Phase n|],[|Phase (n-1)|])
      in add_phase_rules (r::rules) (n-1)
;;

let get_all_rules twoConstants max_tuple_size parsed_protocol parsed_facts fresh_var =
        let rec ptcl_rules maxPh rule_list input_types fact_list fresh_id = function
        | [] -> maxPh,rule_list,input_types,fact_list
        | (p,q)::t -> 
                        let maxPh1,new_rule_list,new_input_types,_,nfid  =  to_rule maxPh p q rule_list input_types fresh_var fresh_id in
                        let p_null = is_null p in
                        let q_null = is_null q in
                        begin
                        match p_null,q_null with
                        | false,_     -> 
                                        let f = State(get_channel p,0,[||],[||]) 
                                        in ptcl_rules maxPh1 new_rule_list new_input_types (f::fact_list) nfid t
                        | true,false ->  
                                        let f = State(get_channel q,0,[||],[||]) 
                                        in ptcl_rules maxPh1 new_rule_list new_input_types (f::fact_list) nfid t 
                        | true,true -> ptcl_rules maxPh1 new_rule_list new_input_types fact_list fresh_id t 
                        end
                        
        in let maxPh,rule_list,_,fact_list = ptcl_rules 0 (decomposition_rules max_tuple_size) [] parsed_facts 0 parsed_protocol
        in let minimal_facts = (Ik(At(Cst(0),(At(Cameleon))),c 0)) :: (Phase 0)::fact_list
        in let all_facts = match twoConstants with
        |true  -> Ik(At(Cst(1),(At(Cameleon))) ,c 1) ::  minimal_facts
        |false -> minimal_facts
        in all_facts,add_phase_rules rule_list maxPh
;;
